/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AttachmentCount.
 *
 * The Initial Developer of the Original Code is
 *   alta88 <alta88@gmail.com>.
 *
 * Portions created by the Initial Developer are Copyright (C) 2015
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

var {classes: Cc, interfaces: Ci, utils: Cu} = Components;

Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/NetUtil.jsm");
Cu.import("resource:///modules/gloda/log4moz.js");
Cu.import("resource:///modules/StringBundle.js");

var AttachmentCount = {
  prefBranch: "extensions.AttachmentCount",

  d: function(elementId)   { return document.getElementById(elementId); },

  get log() {
    return Log4Moz.getConfiguredLogger(this.prefBranch);
  },

  get strings() {
    return this._strings ? this._strings :
      this._strings = new StringBundle("chrome://attachmentcount/locale/ac.properties");
  },

  get attachmentCountCol() {
    return GetThreadTree().columns["attachmentCountCol"];
  },

  get alwaysShowCount() {
    return Services.prefs.getBoolPref(this.prefBranch + ".alwaysShowCount");
  },

  get showAllPartsPref() {
    return Services.prefs.getIntPref("mailnews.display.html_as") == 4;
  },

  get downloadPrefs() {
    return JSON.parse(Services.prefs.getCharPref(this.prefBranch + ".downloadPrefs"));
  },
  get imapAllowDownload()  {return this.downloadPrefs.imapAllowDownload;},
  get saneBodySize()       {return this.downloadPrefs.saneBodySize;},
  get imapPartsOnDemand()  {return this.downloadPrefs.imapPartsOnDemand;},
  // Throws if false for nntp.
  get encryptedParts()     {return true;},

  get threadTreeChildren() {
    return GetThreadTree().getElementsByTagName("treechildren")[0];
  },
  get previewEnabled() {
    return Services.prefs.getBoolPref(this.prefBranch + ".previewEnabled");
  },
  get previewContentTypes() {
    return Services.prefs.getCharPref(this.prefBranch + ".previewContentTypes");
  },
  get previewTitle() {
    return this.strings.get("attachmentPreviewTitle");
  },
  get previewTitleMulti() {
    return this.strings.get("attachmentPreviewTitleMulti");
  },
  get previewTitleNA() {
    return this.strings.get("attachmentPreviewTitleNA");
  },
  previewFilename: function (aName) {
    return this.strings.get("attachmentPreviewFilename", [aName]);
  },
  previewContentType: function (aContentType) {
    return this.strings.get("attachmentPreviewContentType", [aContentType]);
  },
  previewDetachedFile: function (aFile) {
    return this.strings.get("attachmentPreviewDetachedFile", [aFile]);
  },
  previewLink: function (aUrl) {
    return this.strings.get("attachmentPreviewLink", [aUrl]);
  },
  get previewSeparator() {
    return this.strings.get("attachmentPreviewSeparator");
  },

  onLoad: function () {
    Services.obs.addObserver(AttachmentCount.Observer, "MsgCreateDBView", false);
    Services.prefs.addObserver("mailnews.display.html_as", AttachmentCount.Observer, false);
//    AttachmentCount._TreeOnMouseDown = TreeOnMouseDown;
//    TreeOnMouseDown = AttachmentCount.TreeOnMouseDown;
//    GetThreadTree().addEventListener("click",ThreadPaneOnClick,true);
//    TM.u.threadTreeChildren.addEventListener("click", MultiRow.onClick, true);
    AttachmentCount.threadTreeChildren.addEventListener("mousemove", AttachmentCount.onMouseMove, false);
    AttachmentCount.threadTreeChildren.addEventListener("mouseleave", AttachmentCount.onMouseLeave, false);
  },

  TreeOnMouseDown: function (aEvent) {
    if (aEvent.target.parentNode != GetThreadTree())
      return AttachmentCount._TreeOnMouseDown(aEvent);
AttachmentCount.log.debug("TreeOnMouseDown: START name - "+aEvent.target.localName);
    let cell = GetThreadTree().treeBoxObject.getCellAt(aEvent.clientX, aEvent.clientY);
    let row = cell.row;
    let col = cell.col;
    if (row < 0 || aEvent.button != 0 ||
        aEvent.target.localName == "treecol" ||
        (col && col.id != "attachmentCountCol"))
      return AttachmentCount._TreeOnMouseDown(aEvent);

AttachmentCount.log.debug("TreeOnMouseDown: CONTINUE row:msgKey - "+
                          row+":"+gDBView.getMsgHdrAt(row).messageKey);
    aEvent.stopPropagation();
    let msgHdr = gDBView.getMsgHdrAt(row);
    AttachmentCount.threadTreeChildren.removeAttribute("tooltiptext");
    AttachmentCount.setInfo(msgHdr, "attachmentPreview", undefined);
//    AttachmentCount.setCountProperty(msgHdr, null);
    AttachmentCount.resetOne = true;
    GetThreadTree().treeBoxObject.invalidateCell(row, col);
//    AttachmentCount.reset = false;
    return true;
  },

  ThreadPaneOnClick: function (aEvent) {
AttachmentCount.log.debug("ThreadPaneOnClick: START id - "+aEvent.target.id);
    if (aEvent.shiftKey && aEvent.target.id == "attachmentCountCol")
      return AttachmentCount.toggleReset(aEvent);

    return AttachmentCount._ThreadPaneOnClick(aEvent);
  },

  onMouseMove: function (aEvent) {
//AttachmentCount.log.debug("onMouseMove: START name - "+aEvent.target.localName);
    if (AttachmentCount.gettingPreview || !AttachmentCount.previewEnabled)
      return;

    AttachmentCount.gettingPreview = false;
    let cell = GetThreadTree().treeBoxObject.getCellAt(aEvent.clientX, aEvent.clientY);
    let row = cell.row;
    let col = cell.col;

    let previewRow = AttachmentCount.previewRow < 0 ||
                     AttachmentCount.previewRow >= gDBView.rowCount ?
                       -1 : AttachmentCount.previewRow;
    if (row < 0 || col.id != "attachmentCountCol") {
      AttachmentCount.threadTreeChildren.removeAttribute("tooltiptext");
      AttachmentCount.previewRow = null;
      GetThreadTree().treeBoxObject.invalidateCell(previewRow, AttachmentCount.attachmentCountCol);
      return;
    }

    let msgHdr = gDBView.getMsgHdrAt(row);
    if (AttachmentCount.previewMsgHdr && msgHdr != AttachmentCount.previewMsgHdr) {
      AttachmentCount.threadTreeChildren.removeAttribute("tooltiptext");
      AttachmentCount.previewRow = null;
      GetThreadTree().treeBoxObject.invalidateCell(previewRow, col);
      GetThreadTree().treeBoxObject.invalidateCell(row, col);
    }

//AttachmentCount.log.debug("onMouseMove: NEXT msgKey:previewMsgKey - "+
//                          msgHdr.messageKey+":"+
//                          (AttachmentCount.previewMsgHdr ? AttachmentCount.previewMsgHdr.messageKey : null));
    if (!AttachmentCount.isPreviewValid(msgHdr) ||
        AttachmentCount.threadTreeChildren.hasAttribute("tooltiptext")) {
      return;
    }

    let preview = AttachmentCount.getInfo(msgHdr, "attachmentPreview");
    if (preview === null) {
//AttachmentCount.log.debug("onMouseMove: NO PREVIEW msgKey - "+msgHdr.messageKey);
      return;
    }
    else if (preview == undefined) {
      AttachmentCount.gettingPreview = true;
      AttachmentCount.setCountProperty(msgHdr, null);
//AttachmentCount.log.debug("onMouseMove: NEW RESET msgKey - "+msgHdr.messageKey);
    }
    else {
      AttachmentCount.threadTreeChildren.setAttribute("tooltiptext", preview);
      AttachmentCount.previewMsgHdr = msgHdr;
      AttachmentCount.previewRow = row;
//AttachmentCount.log.debug("onMouseMove: NEW CACHED msgKey - "+msgHdr.messageKey);
    }

    GetThreadTree().treeBoxObject.invalidateCell(previewRow, col);
    GetThreadTree().treeBoxObject.invalidateCell(row, col);
  },

  onMouseLeave: function (aEvent) {
//AttachmentCount.log.debug("onMouseLeave: START id - "+aEvent.target.id);
    AttachmentCount.threadTreeChildren.removeAttribute("tooltiptext");
    let row = AttachmentCount.previewRow;
    AttachmentCount.previewRow = null;
    GetThreadTree().treeBoxObject.invalidateCell(row, AttachmentCount.attachmentCountCol);
  },

  /**
   * nsITreeview.
   */
  isString:            function()         {return true;},
  isEditable:          function(row, col) {return false;},
  cycleCell:           function(row, col) {},
  getRowProperties:    function(row)      {return "";},
  getSortLongForRow:   function(msgHdr)   {return 0;},

  getSortStringForRow: function(msgHdr)   {
    let sortStr = "";
    let ext = ".";
    let countInfo = this.getCount(msgHdr);
    if (!countInfo || countInfo[this.CountInfo.kAttachments] == 0) {
      sortStr = "*000" + ext;
this.log.debug("getSortStringForRow: sortStr:ext:countInfo:msgKey - "+
               sortStr+":::"+msgHdr.messageKey);
      return sortStr;
    }

    if (countInfo[this.CountInfo.kAttachments] == 1 &&
        countInfo[this.CountInfo.kNameExt]) {
      // Get the file ext.  We want to group same type 1 count items
      // together.
      ext += countInfo[this.CountInfo.kNameExt];
    }

    let countStr = countInfo[this.CountInfo.kAttachments].toString();
    while (countStr.length < 3)
      countStr = "0" + countStr;

    // Group all allDeleted/allDetached together.
    let flag = this.kNormal;
    if ([this.kDeleted, this.kDetached].indexOf(countInfo[this.CountInfo.kFlag]) != -1)
      flag = countInfo[this.CountInfo.kFlag];

    sortStr = flag + countStr + ext;
this.log.debug("getSortStringForRow: sortStr:ext:countInfo:msgKey - "+
               sortStr+":"+ext+":"+countInfo+":"+msgHdr.messageKey);
    return sortStr;
  },

  getImageSrc: function(row, col) {
    if (row < 0 || row >= gDBView.rowCount || this.isGroupedBySortHdr(row) ||
        this.showAllPartsPref || this.alwaysShowCount)
      return null;

    let countInfo = this.getCount(gDBView.getMsgHdrAt(row));
    if (!countInfo || countInfo[this.CountInfo.kAttachments] != 1 ||
        [this.kDeleted, this.kDetached].
         indexOf(countInfo[this.CountInfo.kFlag]) != -1)
      return null;

    let iconStr = "moz-icon://sheep?size=16";
    if (!countInfo[this.CountInfo.kNameExt])
      return iconStr;

    let contentTypeStr = "";
    let contentType = countInfo[this.CountInfo.kContentType];
    if (contentType && contentType != "null")
      contentTypeStr = "&contentType=" + countInfo[this.CountInfo.kContentType];
    iconStr = "moz-icon://lamb." + countInfo[this.CountInfo.kNameExt] +
              "?size=16" + contentTypeStr;

this.log.trace("getImageSrc: icon - "+iconStr);
    return iconStr;
  },

  getCellProperties: function(row, col) {
//this.log.debug("getCellProperties: col - "+col.id);
    if (row < 0 || row >= gDBView.rowCount || this.isGroupedBySortHdr(row))
      return null;

    let props = [];
    props.push("attachmentCountCol");
    if (this.alwaysShowCount)
      return props.join(" ");

    let countInfo = this.getCount(gDBView.getMsgHdrAt(row));
    if (!countInfo)
      return props.join(" ");

this.log.trace("getCellProperties: count - "+countInfo);
    if (countInfo[this.CountInfo.kAttachments] == 1 ||
        [this.kDeleted, this.kDetached].indexOf(countInfo[this.CountInfo.kFlag]) != -1)
      props.push("hasOne");
    if (countInfo[this.CountInfo.kAttachments] > 1 &&
        [this.kDeleted, this.kDetached].indexOf(countInfo[this.CountInfo.kFlag]) == -1)
      props.push("hasMany");
    if (countInfo[this.CountInfo.kAttachments] <= 9)
      props.push("hasMany9");
    if (countInfo[this.CountInfo.kFlag] == this.kDeleted)
      props.push("allDeleted");
    if (countInfo[this.CountInfo.kFlag] == this.kDetached)
      props.push("allDetached");
    if (countInfo[this.CountInfo.kFlag] == this.kRemote)
      props.push("allRemote");
    if (countInfo[this.CountInfo.kFlag] == this.kUnknown)
      props.push("allUnknown");
    if (countInfo[this.CountInfo.kFlag] == this.kUnknownNNTP)
      props.push("allUnknownNNTP");
    if (row == this.previewRow)
      props.push("previewing");

    return props.join(" ");
  },

  getCellText: function(row, col) {
    if (row < 0 || row >= gDBView.rowCount || this.isGroupedBySortHdr(row))
      return null;

    let countInfo = this.getCount(gDBView.getMsgHdrAt(row));
    // Don't display ? for unknown, it is almost surely a case of no attachments.
    // if (countInfo && countInfo[this.CountInfo.kFlag] == this.kUnknown)
    //   return "?";
    if (countInfo && countInfo[this.CountInfo.kFlag] == this.kNoAccess)
      return "?";

    if (!countInfo || !countInfo[this.CountInfo.kAllBodyParts] ||
        countInfo[this.CountInfo.kAttachments] == 0)
      return null;

    let counts = this.showAllPartsPref ? countInfo[this.CountInfo.kAllBodyParts] :
                                         countInfo[this.CountInfo.kAttachments];
//    counts = counts == 18 ? 555 : counts;
    return counts > 99 ? "99+" : counts;
  },

  isGroupedBySortHdr: function(row) {
    return gFolderDisplay.view.showGroupedBySort && row >= 0 &&
           (gDBView.isContainer(row) || gDBView.getLevel(row) == 0);
  }, // nsITreeview.

  AttachmentInfo: {
    kName:         0,
    kContentType:  1,
    kUrl:          2,
    kFlag:         3
  },
  CountInfo: {
    kAttachments:  0,
    kAllBodyParts: 1,
    kFlag:         2,
    kContentType:  3,
    kNameExt:      4
  },
  // kFlag values.
  kDeleted:        0,
  kDetached:       1,
  kRemote:         2,
  kNormal:         3,
  kUnknown:        4,
  kUnknownNNTP:    5,
  kNoAccess:       6,
  get CountInfoLength() {
    return Object.keys(this.CountInfo).length;
  },

  // A shift-click on the attachment count column header will enter reset mode,
  // any message that has been scrolled into view will be reset on a second
  // shift-click that exits reset mode.  The attachmentCounts property will
  // be derived again from the mime message.
  reset: false,
  toggleReset: function(aEvent) {
    AttachmentCount.reset = !AttachmentCount.reset;
    GetThreadTree().treeBoxObject.invalidateColumn(AttachmentCount.attachmentCountCol);
  },
  // A click on the column row will reset the property for a single message.
  resetOne: false,

  gettingCount: false,
  getCountProperty: function(aMsgHdr)  {
    let counts = aMsgHdr.getProperty("attachmentCounts");
this.log.trace("getCountProperty: prop:subject - "+counts+":"+aMsgHdr.mime2DecodedSubject);
    if (counts.length == 0)
      return null;
    counts = counts.split(",");
    counts[0] = parseInt(counts[0]);
    counts[1] = parseInt(counts[1]);
    counts[2] = parseInt(counts[2]);
    return counts;
  },
  setCountProperty:    function(aMsgHdr, aVal)  {
    aMsgHdr.setProperty("attachmentCounts", aVal);
  },
  getCount: function(aMsgHdr)  {
//this.log.debug("getCount: START reset:resetOne:msgKey - "+this.reset+":"+this.resetOne+":"+aMsgHdr.messageKey);
    if (!(aMsgHdr instanceof Ci.nsIMsgDBHdr) ||
        (!(aMsgHdr.flags & Ci.nsMsgMessageFlags.Attachment) &&
         !this.reset && !this.resetOne)) {
      // No msgHdr or attachments.
//this.log.debug("getCount: no msgHdr or attachemnts, msgKey - "+aMsgHdr.messageKey);
      return null;
    }

    if (this.reset) {
      this.setCountProperty(aMsgHdr, null);
      return null;
    }

    let countInfo = this.getCountProperty(aMsgHdr);
    if (countInfo)
      // Return the already stored attachmentCount property.
      return this.lastCountInfo = countInfo;

//this.log.debug("getCount: count:subject - "+countInfo+":"+msgHdr.mime2DecodedSubject);
    // TODO: check if logged in.
    let imapLoggedIn = true;

    if (aMsgHdr.folder.server.type == "imap" &&
        (!this.imapAllowDownload || !imapLoggedIn)) {
      // If there is no count already then not allowing download throws
      // in MsgHdrToMimeMessage, and if not signed in can't get a count.
      countInfo = new Array(AttachmentCount.CountInfoLength),
      countInfo[AttachmentCount.CountInfo.kFlag] = AttachmentCount.kNoAccess;
      return countInfo;
    }

    if (this.gettingCount)
      return null;
    this.gettingCount = true;

    // Get the count from message property attachmentCount.  If not there,
    // derive from the mime message and set it.
    // XXX: MsgHdrToMimeMessage used to be crap.  Thanks to protz for
    // fixing the api.  However, it still chokes on nntp.
this.log.trace("getCount: prefs: "+this.downloadPrefs.toSource());
    MsgHdrToMimeMessage(aMsgHdr, null, this.mimeMsgCallback,
                        this.imapAllowDownload,
                        { saneBodySize:          this.saneBodySize,
                          partsOnDemand:         this.imapPartsOnDemand,
                          examineEncryptedParts: this.encryptedParts });

    // Wait for the async call to finish; return null values.
    return null;
  },

  getInfo: function(aMsgHdr, aInfo) {
    let msgKey = aMsgHdr.messageKey;
    let folderKey = aMsgHdr.folder.URI;
    if (!this._cache[folderKey])
      this._cache[folderKey] = {};
    if (!this._cache[folderKey][msgKey])
      this._cache[folderKey][msgKey] = {};
    if (this._cache[folderKey][msgKey][aInfo] === undefined)
      this._cache[folderKey][msgKey][aInfo] = undefined;

    return this._cache[folderKey][msgKey][aInfo];
  },
  setInfo: function(aMsgHdr, aInfo, aVal) {
    let msgKey = aMsgHdr.messageKey;
    let folderKey = aMsgHdr.folder.URI;
    if (!this._cache[folderKey])
      this._cache[folderKey] = {};
    if (!this._cache[folderKey][msgKey])
      this._cache[folderKey][msgKey] = {};

    this._cache[folderKey][msgKey][aInfo] = aVal;
  },

  MsgSortByAttachmentCount: function() {
    if (!gDBView || !gDBView.db)
      return;

    gDBView.db.dBFolderInfo.setProperty("customSortCol", "attachmentCountCol")
    MsgSortThreadPane("byCustom");
  },

  // Preview.
  _cache: {},
  previewRow: null,
  gettingPreview: false,
  previewAttachments: [],

  isPreviewValid: function(aMsgHdr) {
    let countInfo = this.getCount(aMsgHdr);
    return Boolean(countInfo && countInfo[this.CountInfo.kAttachments] > 0);
  },

  isPreviewContentTypeValid: function (aContentType) {
    if (!aContentType)
      return false;

    if (aContentType.startsWith("text/") || aContentType.startsWith("message/") ||
        aContentType.startsWith("multipart/") ||
        (aContentType.startsWith("application/") &&
         (aContentType.endsWith("ics") || aContentType.endsWith("javascript") ||
          aContentType.endsWith("json") || aContentType.endsWith("rdf") ||
          aContentType.endsWith("xml"))) ||
        this.previewContentTypes.indexOf(aContentType) != -1)
      return true;

    return false;
  },

  preview: function(aMsgHdr) {
this.log.debug("preview: START aMsgHdr.messageKey - "+aMsgHdr.messageKey);
    if (this.previewAttachments.length == 0) {
this.log.debug("preview: PREVIEW, none msgKey - "+aMsgHdr.messageKey);
      this.gettingPreview = false;
      return;
    }

    let name, contentType, url, title, preview;
    if (this.previewAttachments.length > 1) {
      preview = title = this.previewTitleMulti + "\n\n";
      for (let attachmentInfo of this.previewAttachments) {
        name = attachmentInfo[this.AttachmentInfo.kName];
        url = decodeURI(attachmentInfo[this.AttachmentInfo.kUrl]);
        contentType = attachmentInfo[this.AttachmentInfo.kContentType];
        preview = preview + this.previewSeparator + "\n" +
                            this.previewFilename(name) + "\n" +
                            this.previewContentType(contentType) + "\n";
        if (attachmentInfo[this.AttachmentInfo.kFlag] == this.kDetached)
          preview = preview + this.previewDetachedFile(url) + "\n";
        if (attachmentInfo[this.AttachmentInfo.kFlag] == this.kDeleted)
          preview = preview + name + "\n";
        if (attachmentInfo[this.AttachmentInfo.kFlag] == this.kRemote) 
          preview = preview + this.previewLink(url) + "\n";

this.log.debug("preview: PREVIEW, msgKey:info - "+
               aMsgHdr.messageKey + ":" +attachmentInfo);
      }
      this.setInfo(aMsgHdr, "attachmentPreview", preview);
      this.gettingPreview = false;
      return;
    }

    let attachmentInfo = this.previewAttachments[0];
    name = attachmentInfo[this.AttachmentInfo.kName];
    contentType = attachmentInfo[this.AttachmentInfo.kContentType];
    url = decodeURI(attachmentInfo[this.AttachmentInfo.kUrl]);
    if (!this.isPreviewContentTypeValid(contentType) || !url ||
        [this.kRemove, this.kDetached].indexOf(attachmentInfo[this.AttachmentInfo.kFlag]) != -1) {
      title = this.previewTitleNA + "\n\n";
      preview = title + this.previewFilename(name) + "\n" +
                        this.previewContentType(contentType) + "\n";
      if (attachmentInfo[this.AttachmentInfo.kFlag] == this.kDetached)
        preview = preview + this.previewDetachedFile(url) + "\n";
      if (attachmentInfo[this.AttachmentInfo.kFlag] == this.kRemote) 
        preview = preview + this.previewLink(url) + "\n";

      this.setInfo(aMsgHdr, "attachmentPreview", preview);
this.log.debug("preview: INFO ONLY, msgKey:attachmentPreview - "+
               aMsgHdr.messageKey + "\n" +
               this.getInfo(aMsgHdr, "attachmentPreview"));
      this.gettingPreview = false;
      return;
    }
this.log.debug("preview: PREVIEW, msgKey:info - "+
               aMsgHdr.messageKey + ":" +this.previewAttachments[0]);

    let uri = Services.io.newURI(url, null, null);
    let channel = Services.io.newChannelFromURI(uri);
    channel.asyncOpen(this.StreamListener, aMsgHdr);
  },

  // Observer for custom column and show allBodyParts pref.
  Observer: {
    observe: function(aSubject, aTopic, aData) {
      if (aTopic == "MsgCreateDBView" && gDBView && ("AttachmentCount" in window)) {
AttachmentCount.log.debug("Observer: addColumnHandler - attachmentCountCol");
        gDBView.addColumnHandler("attachmentCountCol", AttachmentCount);
        gFolderDisplay.view.dbView.db.AddListener(AttachmentCount.DBChangeListener);
        AttachmentCount.reset = false;
      }
      else if (aTopic == "nsPref:changed" && aData == "mailnews.display.html_as")
        GetThreadTree().treeBoxObject.invalidateColumn(AttachmentCount.attachmentCountCol);
    }
  },

  // Listener for header re-add on attachment detach/delete.
  DBChangeListener: {
    onHdrFlagsChanged:    function(aHdrChanged, aOldFlags, aNewFlags, aInstigator)   {},
    onHdrDeleted:         function(aHdrChanged, aParentKey, aFlags, aInstigator)     {},
    onHdrPropertyChanged: function(aHdrToChange, aPreChange, aStatus, aInstigator)   {},
    onParentChanged:      function(aKeyChanged, aOldParent, aNewParent, aInstigator) {},
    onAnnouncerGoingAway: function(aInstigator) {},
    onReadChanged:        function(aInstigator) {},
    onJunkScoreChanged:   function(aInstigator) {},
    onEvent:              function(aDB, aEvent) {},
    // On attachment detach/delete, clear property to recreate new state.
    onHdrAdded:           function(aHdrChanged, aParentKey, aFlags, aInstigator) {
//      gFolderDisplay.view.dbView.db.RemoveListener(AttachmentCount.Listener);
      AttachmentCount.setCountProperty(aHdrChanged, null);
AttachmentCount.log.debug("DBChangeListener.onHdrAdded: attachmentCount reset - "+aHdrChanged.messageKey);
    }
  },

  // Listener for streaming attachment.
  StreamListener: {
    onStartRequest: function (aRequest, aContext) {
      this.msgHdr = aContext;
      this.unicodeConverter = Cc["@mozilla.org/intl/scriptableunicodeconverter"]
                                .createInstance(Ci.nsIScriptableUnicodeConverter);
      this.unicodeConverter.charset = "UTF-8";
      this.data = "";
      this.maxLimit = 4000;
    },
    onDataAvailable: function (aRequest, aContext, aStream, aOffset, aCount) {
AttachmentCount.log.trace("StreamListener.onDataAvailable: aCount - "+aCount);
      let bytes = NetUtil.readInputStreamToString(aStream, aCount);
      if (this.data.length >= this.maxLimit)
        // Don't construct a data buffer bigger than the maxLimit. Is there a
        // way to not read the whole stream..
        return;

      let byteCount = this.maxLimit - this.data.length;
      byteCount = byteCount > aCount ? aCount : byteCount;
AttachmentCount.log.trace("StreamListener.onDataAvailable: data length - "+ this.data.length);

      // Each character of the string is a byte of a UTF-8 string. Use charCodeAt.
      let array = [];
      for (let i = 0; i < byteCount; ++i)
        array[i] = bytes.charCodeAt(i);
AttachmentCount.log.trace("StreamListener.onDataAvailable: array length - "+ array.length);

      this.data += this.unicodeConverter.convertFromByteArray(array, array.length);

AttachmentCount.log.trace("StreamListener.onDataAvailable: +=data length - "+ this.data.length);
    },
    onStopRequest: function (aRequest, aContext, aStatusCode) {
AttachmentCount.log.debug("StreamListener.onStopRequest: final data len - "+ this.data.length);
      let self = AttachmentCount;
      let title = self.previewTitle + "\n\n";
      let name = self.previewAttachments[0][self.AttachmentInfo.kName];
      let contentType = self.previewAttachments[0][self.AttachmentInfo.kContentType];
      let preview = title + self.previewFilename(name) + "\n" +
                            self.previewContentType(contentType) + "\n";
      if (self.previewAttachments[0][self.AttachmentInfo.kFlag] == self.kDeleted)
        preview = preview + name + "\n";
      preview = preview + self.previewSeparator + "\n" +
                          this.data + "\n";
      self.setInfo(this.msgHdr, "attachmentPreview", preview);
AttachmentCount.log.debug("StreamListener.onStopRequest: msgKey:attachmentPreview - "+
                          this.msgHdr.messageKey + "\n" +
                          self.getInfo(this.msgHdr, "attachmentPreview"));
      self.gettingPreview = false;
    }
  },

  mimeMsgCallback: function(aMsgHdr, aMimeMsg) {
    let part, val, partname, partcontentType,
        count = {c:0, m:0, b:0, a:0, u:0},
        countInfo = new Array(AttachmentCount.CountInfoLength),
        deletedCount = 0,
        detachedCount = 0,
        remoteCount = 0,
        msgHdr = aMsgHdr;
AttachmentCount.log.debug("mimeMsgCallback: aMsgKey - "+aMsgHdr.messageKey);

    if (aMimeMsg == null) {
AttachmentCount.log.debug("mimeMsgCallback: aMimeMsg is null ");
      // Couldn't get a mime message.  Always true for nntp.
      if (msgHdr.folder.server.type == "nntp")
        countInfo[AttachmentCount.CountInfo.kFlag] = AttachmentCount.kUnknownNNTP;
      else
        countInfo[AttachmentCount.CountInfo.kFlag] = AttachmentCount.kUnknown;
      AttachmentCount.setCountProperty(msgHdr, countInfo);
      AttachmentCount.gettingCount = false;
      GetThreadTree().treeBoxObject.invalidateColumn(AttachmentCount.attachmentCountCol);
      return;
    }

AttachmentCount.log.debug("mimeMsgCallback: aMimeMsg: "+aMimeMsg.prettyString());
AttachmentCount.log.trace("mimeMsgCallback: aMimeMsg: "+aMimeMsg.toSource());

    // Implement recursive parser ourselves..
    function counter(parts) {
      for (let [index, part] in Iterator(parts)) {

if (!AttachmentCount.reset)
AttachmentCount.log.debug("counter: part - "+part.toSource());

        if (part instanceof MimeContainer) {
          count.c++;
if (!AttachmentCount.reset)
AttachmentCount.log.debug("mimeMsgCallback.counter: MimeContainer " +
                          "partName:count:contentType:name - "+
                          part.partName+" : "+count.c+" : "+
                          part.contentType+" : "+part.name);
        }

        if (part instanceof MimeMessage) {
          count.m++;
          partname = part.name ? part.name.match(/\./g)
                               ? part.name : part.name + ".1" : ".1";
          partcontentType = part.contentType;

          let attachmentInfo = [part.name, part.contentType, part.url, AttachmentCount.kNormal];
          AttachmentCount.previewAttachments.push(attachmentInfo);

if (!AttachmentCount.reset)
AttachmentCount.log.debug("mimeMsgCallback.counter: MimeMessage " +
                          "partName:count:contentType:name - "+
                          part.partName+" : "+count.m+" : "+
                          part.contentType+" : "+partname);
        }

        if (part instanceof MimeBody) {
          count.b++;
if (!AttachmentCount.reset)
AttachmentCount.log.debug("mimeMsgCallback.counter: MimeBody " +
                          "partName:count:contentType:name - "+
                          part.partName+" : "+count.b+" : "+
                          part.contentType+" : "+part.name);
        }

        if (part instanceof MimeMessageAttachment) {
          count.a++;
          // Use a ".1" extension if none or no name, to show a blank icon.
          partname = part.name ? part.name.match(/\./g)
                               ? part.name : part.name + ".1" : ".1";

          if (part.contentType == "multipart/appledouble")
            // Handle appledouble part types, contains 2 children, the 2nd of
            // which has the real content type.
            partcontentType = part.parts[1].contentType;
          else
            partcontentType = part.contentType;

          let flag = AttachmentCount.kNormal;
          // Deleted attachments.
          if (partcontentType == "text/x-moz-deleted") {
            deletedCount++;
            flag = AttachmentCount.kDeleted;
          }

          // Detached attachments and Feed enclosures (isExternal=true)
          // are correctly returned as MimeMessageAttachment in Tb8.0+.
          // Differentiate between them.
          let partUrl = part.url;
          if (part.isExternal) {
AttachmentCount.log.debug("mimeMsgCallback.counter: MimeMessageAttachment " +
                          "part.url:"+part.url);
            if (part.url.match(/^(http|ftp)/)) {
              // A feed message url is 'remote', http or ftp.  A feed
              // messge doesn't have to be in a feed folder.
              remoteCount++;
              flag = AttachmentCount.kRemote;
            }
            else {
              detachedCount++;
              flag = AttachmentCount.kDetached;
            }
            partUrl = part.url.replace(/[?&]part=.+$/, "");
          }

          let attachmentInfo = [part.name, partcontentType, partUrl, flag];
          AttachmentCount.previewAttachments.push(attachmentInfo);

if (!AttachmentCount.reset)
AttachmentCount.log.debug("mimeMsgCallback.counter: MimeMessageAttachment " +
                          "partName:count:contentType:name - "+
                          part.partName+" : "+count.a+" : "+
                          partcontentType+" : "+partname);
        }

        if (part instanceof MimeUnknown) {
          count.u++;
if (!AttachmentCount.reset) {
AttachmentCount.log.debug("mimeMsgCallback.counter: MimeUnknown " +
                          "part headers.content-type: "+
                          part.headers["content-type"][0]);
AttachmentCount.log.debug("mimeMsgCallback.counter: MimeUnknown " +
                          "partName:count:contentType:name - "+
                          part.partName+" : "+count.u+" : "+
                          part.contentType+" : "+part.name);
}
        }

        if (part.parts && part.contentType != "multipart/appledouble")
          counter(part.parts);

      } // for
    } // counter()

    AttachmentCount.previewAttachments = [];
    counter(aMimeMsg.parts);
    let attachments = count.m + count.a;
    let allBodyParts = attachments + count.b + count.u;
    countInfo[AttachmentCount.CountInfo.kAttachments] = attachments;
    countInfo[AttachmentCount.CountInfo.kAllBodyParts] = allBodyParts;
    if (attachments == 0 && !AttachmentCount.reset && !AttachmentCount.resetOne) {
      // The msgHdr says we have an attachment but we didn't count any.
      // For multipart messages, the flag is preset to true, and updated only
      // when the message is opened (leading to false indicators until then).
      //
      // Or, there may be an error in parsing, or there was no 'name' in
      // content-type test; set unknown indicator and log a warning.
      countInfo[AttachmentCount.CountInfo.kFlag] = AttachmentCount.kUnknown;
AttachmentCount.log.warn("mimeMsgCallback: MimeUnknown couldn't get a count " +
                         "though nsMsgMessageFlags.Attachment is set");
    }
    // All of a certain type get their own decoration..
    else if (attachments == deletedCount)
      countInfo[AttachmentCount.CountInfo.kFlag] = AttachmentCount.kDeleted;
    else if (attachments == detachedCount)
      countInfo[AttachmentCount.CountInfo.kFlag] = AttachmentCount.kDetached;
    else if (attachments == remoteCount)
      countInfo[AttachmentCount.CountInfo.kFlag] = AttachmentCount.kRemote;
    else
      countInfo[AttachmentCount.CountInfo.kFlag] = AttachmentCount.kNormal;

    if (attachments == 1) {
      countInfo[AttachmentCount.CountInfo.kContentType] = partcontentType;
      let ext = partname.split(".");
      ext = ext[ext.length - 1].toLowerCase();
      countInfo[AttachmentCount.CountInfo.kNameExt] = ext;
AttachmentCount.log.debug("mimeMsgCallback: partname:kNameExt - "+
                          partname+":"+countInfo[AttachmentCount.CountInfo.kNameExt]);
    }

    AttachmentCount.setCountProperty(msgHdr, countInfo);
    AttachmentCount.gettingCount = false;
    AttachmentCount.resetOne = false;
    GetThreadTree().treeBoxObject.invalidateColumn(AttachmentCount.attachmentCountCol);

    if (AttachmentCount.gettingPreview)
      AttachmentCount.preview(msgHdr);

if (!AttachmentCount.reset)
AttachmentCount.log.debug("mimeMsgCallback: Finish: countInfo:msgKey:attachmentCountsProp - "+
                          countInfo+":"+msgHdr.messageKey+":"+ AttachmentCount.getCount(msgHdr));

  }
}; // AttachmentCount

/*
AttachmentInfo.prototype.detach = function AttachmentInfo_detach(aSaveFirst) {
AttachmentCount.log.trace("AttachmentInfo.detach: START "+aSaveFirst);
  // The sequence of code flow is async for detach, but not delete, with
  // no way to wait for the final confirm dialog, therefore we use this
  // listener.  Note that on a cancel (no way to wait for that either) the
  // listener will remain until a folder is changed.
  messenger.detachAttachment(this.contentType, this.url,
                             encodeURIComponent(this.name),
                             this.uri, aSaveFirst);
//  gFolderDisplay.view.dbView.db.AddListener(AttachmentCount.DBChangeListener);
AttachmentCount.log.debug("AttachmentInfo.detach: DONE");
};

AttachmentCount._HandleMultipleAttachments = HandleMultipleAttachments;
HandleMultipleAttachments = function(attachments, action) {
AttachmentCount.log.trace("HandleMultipleAttachments: START "+action);
  AttachmentCount._HandleMultipleAttachments(attachments, action);
//  gFolderDisplay.view.dbView.db.AddListener(AttachmentCount.DBChangeListener);
AttachmentCount.log.debug("HandleMultipleAttachments: DONE");
};*/

if ("TreeOnMouseDown" in window) {
  AttachmentCount._TreeOnMouseDown = TreeOnMouseDown;
  TreeOnMouseDown = AttachmentCount.TreeOnMouseDown;
}

if ("ThreadPaneOnClick" in window) {
  AttachmentCount._ThreadPaneOnClick = ThreadPaneOnClick;
  ThreadPaneOnClick = AttachmentCount.ThreadPaneOnClick;
}

window.addEventListener("load", AttachmentCount.onLoad, false);
